/****************************************************************************
 *
 * NAME:   Adeos Automated Filesystem Security Scanner
 * AUTHOR: Michael Lamont <lamonml@linux.wku.edu>
 * DATE:   JUL-25-2001
 *
 * DESCRIPTION:
 * -Conducts a systematic scan of the local filesystem looking for common
 *  security problems.  Three different levels of scanning are available, 
 *  and the output can be formatted as plain text or HTML.
 *
 * USAGE:
 *   adeos [OPTIONS]... [SCANTYPE]
 *
 * Available options:
 *   -r        Format output in a collated report style
 *   -h        Format output as HTML
 *   -d        Include dynamic and device directories in the scan
 *   --help    Display help and usage information
 *
 * Scan levels: normal, verbose, paranoid
 *
 * Example Usage:
 *   adeos -r -h verbose
 *
 * Copyright (C) 2001 Michael Lamont 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The author may be reached electronically at lamonml@linux.wku.edu or
 * by postal mail at 35-8 Briarwood Ln, Marlborough, MA 01752, USA.
 *
 ****************************************************************************/


#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <ftw.h>
#include <string.h>
#include <stdlib.h>

struct Tnode
{
  char path[128];
  struct Tnode *next;
};

void initStorage();
int checkFile(const char*, const struct stat*, int);
void outputReport();
void writeHTMLHeaders();
void writeHTMLFooters();
void outputHTMLReport();



struct Tnode *s_isuid, *s_isgid, *s_iwoth, *ds_iwoth;
struct Tnode *s_isuid_cur, *s_isgid_cur, *s_iwoth_cur, *ds_iwoth_cur;
struct Tnode *s_isvtx, *ds_iroth, *dnr, *ns;
struct Tnode *s_isvtx_cur, *ds_iroth_cur, *dnr_cur, *ns_cur;
struct Tnode *s_iroth, *s_ixoth, *ds_ixoth;
struct Tnode *s_iroth_cur, *s_ixoth_cur, *ds_ixoth_cur;
struct Tnode *temp;
int normal, verbose, paranoid, report, html, dynamic;
FILE *fp;


int main(int argc, char **argv)
{
  int i;

  //initialize customization options
  normal = 0;
  verbose = 0;
  paranoid = 0;
  report = 0;
  html = 0;
  dynamic = 0;

  //if no command line arguments, assume default run
  if (argc <= 1)
    normal = 1;

  //print help/usage info if requested
  if ((argc > 1) && (strcmp(argv[1], "--help") == 0))
  {
    printf("Adeos 1.0, an automated file system security scanning tool.\n");
    printf("Usage: adeos [OPTION]... [SCANTYPE]...\n\n");
    printf("Options:\n");
    printf("  -d          Include dynamic directories in scan (/tmp /dev /proc /devices)\n");
    printf("  -h          Output results as an HTML document\n");
    printf("  -r          Format output as a categorically sorted report\n");
    printf("  --help      Display this help message\n\n");
    printf("Scan Types:\n");
    printf("  normal      Basic security scan (default)\n");
    printf("  verbose     Extended security scan\n");
    printf("  paranoid    Return extended status information for every file\n");
    printf("\nPlease note that this program should be run from a normal\n");
    printf("user account.  The results will be meaningless if it is run\n");
    printf("from a priveleged account such as root.\n");
    printf("\nMail bug reports and suggestions to <lamonml@linux.wku.edu>.\n");
    exit(0);
  }

  //check for command line goodies
  if (argc > 1)
  {
    for (i = 1; i < argc; i++)
    {
      if (strcmp(argv[i], "-h") == 0)
        html = 1;
      else if (strcmp(argv[i], "-r") == 0)
        report = 1;
      else if (strcmp(argv[i], "-d") == 0)
        dynamic = 1;
      else if (strcmp(argv[i], "normal") == 0)
        normal = 1;
      else if (strcmp(argv[i], "verbose") == 0)
        verbose = 1;
      else if (strcmp(argv[i], "paranoid") == 0)
        paranoid = 1;
      else
        {
          printf("adeos: unknown or invalid option\n");
          printf("Usage: adeos [OPTION]... [SCANTYPE]...\n\n");
          printf("Try 'adeos --help' for more information.\n\n");
          exit(1);
        }
    }
  }

  if (report)
    initStorage();

  if (html)
    writeHTMLHeaders();

  nftw("/", checkFile, sysconf(_SC_OPEN_MAX) - 3, 1);

  if (report && html)
    outputHTMLReport();
  else if (report)
    outputReport();

  if (html)
    writeHTMLFooters();

}


int checkFile(const char *path, const struct stat *st, int flag)
{
  //produce no output if we're ignoring /dev, /tmp, and /proc
  if (!dynamic)
  {
    if ((path[0] == '/') && 
        (path[1] == 'd') &&
        (path[2] == 'e') && 
        (path[3] == 'v') && 
        (path[4] == '/'))
      return(0);
    if ((path[0] == '/') &&
        (path[1] == 't') &&
        (path[2] == 'm') &&
        (path[3] == 'p') &&
        (path[4] == '/'))
      return(0);
    if ((path[0] == '/') &&
        (path[1] == 'p') &&
        (path[2] == 'r') &&
        (path[3] == 'o') &&
        (path[4] == 'c') &&
        (path[5] == '/'))
      return(0);
    if ((path[0] == '/') &&
        (path[1] == 'd') &&
        (path[2] == 'e') &&
        (path[3] == 'v') &&
        (path[4] == 'i') &&
        (path[5] == 'c') &&
        (path[6] == 'e') &&
        (path[7] == 's') &&
        (path[8] == '/'))
      return(0);
  }

  switch(flag)
  {
    case FTW_F:
      //normal checks
      if ((st->st_mode & S_ISUID) != 0)
        if (report)
        {
          strcpy(s_isuid_cur->path, path);
          temp = malloc(sizeof(struct Tnode));
          s_isuid_cur->next = temp;
          s_isuid_cur = temp;
          temp->next = NULL;
        }
        else if (html)
        {
          fprintf(fp, "<TR>\n");
          fprintf(fp, "<TD BGCOLOR=red>SUID File</TD>\n");
          fprintf(fp, "<TD>%s</TD\n", path); 
          fprintf(fp, "</TR>\n");
        }
        else
          printf("SUID file: %s\n", path);
      if ((st->st_mode & S_ISGID) != 0)
        if (report)
        {
          strcpy(s_isgid_cur->path, path);
          temp = malloc(sizeof(struct Tnode));
          s_isgid_cur->next = temp;
          s_isgid_cur = temp;
          temp->next = NULL;
        }
        else if (html)
        {
          fprintf(fp, "<TR>\n");
          fprintf(fp, "<TD BGCOLOR=red>SGID File</TD>\n");
          fprintf(fp, "<TD>%s</TD\n", path);
          fprintf(fp, "</TR>\n");
        }
        else
          printf("SGID file: %s\n", path);
      if ((st->st_mode & S_IWOTH) != 0)
        if (report)
        {
          strcpy(s_iwoth_cur->path, path);
          temp = malloc(sizeof(struct Tnode));
          s_iwoth_cur->next = temp;
          s_iwoth_cur = temp;
          temp->next = NULL;
        }
        else if (html)
        {
          fprintf(fp, "<TR>\n");
          fprintf(fp, "<TD BGCOLOR=red>World-Writeable File</TD>\n");
          fprintf(fp, "<TD>%s</TD\n", path);
          fprintf(fp, "</TR>\n");
        }
        else
          printf("World-writeable file: %s\n", path);

      //verbose checks
      if (verbose || paranoid)
      {
        if ((st->st_mode & S_ISVTX) != 0)
          if (report)
          {
            strcpy(s_isvtx_cur->path, path);
            temp = malloc(sizeof(struct Tnode));
            s_isvtx_cur->next = temp;
            s_isvtx_cur = temp;
            temp->next = NULL;
          }
        else if (html)
        {
          fprintf(fp, "<TR>\n");
          fprintf(fp, "<TD BGCOLOR=yellow>\"Sticky\" File</TD>\n");
          fprintf(fp, "<TD>%s</TD\n", path);
          fprintf(fp, "</TR>\n");
        }
          else
            printf("Sticky file: %s\n", path);
      }

     //paranoid checks
     if (paranoid)
     {
        if ((st->st_mode & S_IROTH) != 0)
          if (report)
          {
            strcpy(s_iroth_cur->path, path);
            temp = malloc(sizeof(struct Tnode));
            s_iroth_cur->next = temp;
            s_iroth_cur = temp;
            temp->next = NULL;
          }
        else if (html)
        {
          fprintf(fp, "<TR>\n");
          fprintf(fp, "<TD>World-Readable File</TD>\n");
          fprintf(fp, "<TD>%s</TD\n", path);
          fprintf(fp, "</TR>\n");
        }
          else
            printf("World-readable file: %s\n", path);
       if ((st->st_mode & S_IXOTH) != 0)
         if (report)
         {
           strcpy(s_ixoth_cur->path, path);
           temp = malloc(sizeof(struct Tnode));
           s_ixoth_cur->next = temp;
           s_ixoth_cur = temp;
           temp->next = NULL;
         }
        else if (html)
        {
          fprintf(fp, "<TR>\n");
          fprintf(fp, "<TD>World-Executable File</TD>\n");
          fprintf(fp, "<TD>%s</TD\n", path);
          fprintf(fp, "</TR>\n");
        }
         else
           printf("World-executable file: %s\n", path);
     }
      break;

    case FTW_D:
      //normal checks 
      if ((st->st_mode & S_IWOTH) != 0)
        if (report)
        {
          strcpy(ds_iwoth_cur->path, path);
          temp = malloc(sizeof(struct Tnode));
          ds_iwoth_cur->next = temp;
          ds_iwoth_cur = temp;
          temp->next = NULL;
        }
        else if (html)
        {
          fprintf(fp, "<TR>\n");
          fprintf(fp, "<TD BGCOLOR=yellow>World-Writeable Directory</TD>\n");
          fprintf(fp, "<TD>%s</TD\n", path);
          fprintf(fp, "</TR>\n");
        }
        else
          printf("World-writeable directory: %s\n", path);

      //paranoid checks
      if (paranoid)
      {
        if ((st->st_mode & S_IROTH) != 0)
          if (report)
          {
            strcpy(ds_iroth_cur->path, path);
            temp = malloc(sizeof(struct Tnode));
            ds_iroth_cur->next = temp;
            ds_iroth_cur = temp;
            temp->next = NULL;
          }
        else if (html)
        {
          fprintf(fp, "<TR>\n");
          fprintf(fp, "<TD>World-Readable Directory</TD>\n");
          fprintf(fp, "<TD>%s</TD\n", path);
          fprintf(fp, "</TR>\n");
        }
          else
            printf("World-readable directory: %s\n", path);
        if ((st->st_mode & S_IXOTH) != 0)
          if (report)
          {
            strcpy(ds_ixoth_cur->path, path);
            temp = malloc(sizeof(struct Tnode));
            ds_ixoth_cur->next = temp;
            ds_ixoth_cur = temp;
            temp->next = NULL;
          }
        else if (html)
        {
          fprintf(fp, "<TR>\n");
          fprintf(fp, "<TD>World-Executable Directory</TD>\n");
          fprintf(fp, "<TD>%s</TD\n", path);
          fprintf(fp, "</TR>\n");
        }
          else
            printf("World-executable directory: %s\n", path);
      }
      break;

    case FTW_DNR:
      if (verbose || paranoid)
        if (report)
        {
          strcpy(dnr_cur->path, path);
          temp = malloc(sizeof(struct Tnode));
          dnr_cur->next = temp;
          dnr_cur = temp;
          temp->next = NULL;
        }
        else if (html)
        {
          fprintf(fp, "<TR>\n");
          fprintf(fp, "<TD>Unreadable Directory</TD>\n");
          fprintf(fp, "<TD>%s</TD\n", path);
          fprintf(fp, "</TR>\n");
        }
        else
          printf("Could not read directory: %s\n", path);
      break;

    case FTW_NS:
      if (verbose || paranoid)
        if (report)
        {
          strcpy(ns_cur->path, path);
          temp = malloc(sizeof(struct Tnode));
          ns_cur->next = temp;
          ns_cur = temp;
          temp->next = NULL;
        }
        else if (html)
        {
          fprintf(fp, "<TR>\n");
          fprintf(fp, "<TD BGCOLOR=yellow>Inaccessible File</TD>\n");
          fprintf(fp, "<TD>%s</TD\n", path);
          fprintf(fp, "</TR>\n");
        }
        else
          printf("Could not analyze file: %s\n", path);
      break;  
  }

  return(0);
}


void initStorage()
{
  s_isuid = malloc(sizeof(struct Tnode));
  s_isuid_cur = s_isuid;
  s_isuid->next = NULL;

  s_isgid = malloc(sizeof(struct Tnode));
  s_isgid_cur = s_isgid;
  s_isgid->next = NULL;

  s_iwoth = malloc(sizeof(struct Tnode));
  s_iwoth_cur = s_iwoth;
  s_iwoth->next = NULL;

  ds_iwoth = malloc(sizeof(struct Tnode));
  ds_iwoth_cur = ds_iwoth;
  ds_iwoth->next = NULL;

  if (verbose || paranoid)
  {
    s_isvtx = malloc(sizeof(struct Tnode));
    s_isvtx_cur = s_isvtx;
    s_isvtx->next = NULL;

    dnr = malloc(sizeof(struct Tnode));
    dnr_cur = dnr;
    dnr->next = NULL;

    ns = malloc(sizeof(struct Tnode));
    ns_cur = ns;
    ns->next = NULL;
  }

  if (paranoid)
  {
    ds_iroth = malloc(sizeof(struct Tnode));
    ds_iroth_cur = ds_iroth;
    ds_iroth->next = NULL;

    s_iroth = malloc(sizeof(struct Tnode));
    s_iroth_cur = s_iroth;
    s_iroth->next = NULL;

    s_ixoth = malloc(sizeof(struct Tnode));
    s_ixoth_cur = s_ixoth;
    s_ixoth->next = NULL;

    ds_ixoth = malloc(sizeof(struct Tnode));
    ds_ixoth_cur = ds_ixoth;
    ds_ixoth->next = NULL;
  }
}



void outputReport()
{
  printf("\n");
  printf("                       Results From\n");
  printf("         Adeos Automated Filesystem Security Scanner\n");

  //output SUID file list
  printf("\n\nSUID Files\n");
  printf("-----------------------------------------------------------------\n");
  s_isuid_cur = s_isuid;
  while (s_isuid_cur->next != NULL)
  {
    printf("%s\n", s_isuid_cur->path);
    temp = s_isuid_cur;
    s_isuid_cur = s_isuid_cur->next;
    free(temp);
  }

  //output SGID file list
  printf("\n\nSGID Files\n");
  printf("-----------------------------------------------------------------\n");
  s_isgid_cur = s_isgid;
  while (s_isgid_cur->next != NULL)
  {
    printf("%s\n", s_isgid_cur->path);
    temp = s_isgid_cur;
    s_isgid_cur = s_isgid_cur->next;
    free(temp);
  }

  //output "sticky" files
  if (verbose || paranoid)
  {
    printf("\n\n\"Sticky\" Files\n");
    printf("-----------------------------------------------------------------\n");
    s_isvtx_cur = s_isvtx;
    while (s_isvtx_cur->next != NULL)
    {
      printf("%s\n", s_isvtx_cur->path);
      temp = s_isvtx_cur->next;
      s_isvtx_cur = s_isvtx_cur->next;
      free(temp);
    }
  }

  //output inaccessible files
  if (verbose || paranoid)
  {
    printf("\n\nInaccessible Files\n");
    printf("-----------------------------------------------------------------\n"
);
    ns_cur = ns;
    while (ns_cur->next != NULL)
    {
      printf("%s\n", ns_cur->path);
      temp = ns_cur->next;
      ns_cur = ns_cur->next;
      free(temp);
    }
  }

  //output world-readable files
  if (paranoid)
  {
    printf("\n\nWorld-Readable Files\n");
    printf("-----------------------------------------------------------------\n"
);
    s_iroth_cur = s_iroth;
    while (s_iroth_cur->next != NULL)
    {
      printf("%s\n", s_iroth_cur->path);
      temp = s_iroth_cur->next;
      s_iroth_cur = s_iroth_cur->next;
      free(temp);
    }
  }

  //output world-writeable file list
  printf("\n\nWorld-Writeable Files\n");
  printf("-----------------------------------------------------------------\n");
  s_iwoth_cur = s_iwoth;
  while (s_iwoth_cur->next != NULL)
  {
    printf("%s\n", s_iwoth_cur->path);
    temp = s_iwoth_cur;
    s_iwoth_cur = s_iwoth_cur->next;
    free(temp);
  }

  //output world-executable files
  if (paranoid)
  {
    printf("\n\nWorld-Executable Files\n");
    printf("-----------------------------------------------------------------\n"
);
    s_ixoth_cur = s_ixoth;
    while (s_ixoth_cur->next != NULL)
    {
      printf("%s\n", s_ixoth_cur->path);
      temp = s_ixoth_cur->next;
      s_ixoth_cur = s_ixoth_cur->next;
      free(temp);
    }
  }

  //output world-readable directory list
  if (paranoid)
  {
    printf("\n\nWorld-Readable Directories\n");
    printf("-----------------------------------------------------------------\n"
);
    ds_iroth_cur = ds_iroth;
    while (ds_iroth_cur->next != NULL)
    {
      printf("%s\n", ds_iroth_cur->path);
      temp = ds_iroth_cur->next;
      ds_iroth_cur = ds_iroth_cur->next;
      free(temp);
    }
  }

  //output world-writeable directory list
  printf("\n\nWorld-Writeable Directories\n");
  printf("-----------------------------------------------------------------\n");
  ds_iwoth_cur = ds_iwoth;
  while (ds_iwoth_cur->next != NULL)
  {
    printf("%s\n", ds_iwoth_cur->path);
    temp = ds_iwoth_cur;
    ds_iwoth_cur = ds_iwoth_cur->next;
    free(temp);
  }

  //output world-executable directory list
  if (paranoid)
  {
    printf("\n\nWorld-Executable Directories\n");
    printf("-----------------------------------------------------------------\n");
    ds_ixoth_cur = ds_ixoth;
    while (ds_ixoth_cur->next != NULL)
    {
      printf("%s\n", ds_ixoth_cur->path);
      temp = ds_ixoth_cur;
      ds_ixoth_cur = ds_ixoth_cur->next;
      free(temp);
    }
  }

  //output inaccessible directories
  if (verbose || paranoid)
  {
    printf("\n\nInaccessible Directories\n");
    printf("-----------------------------------------------------------------\n"
);
    dnr_cur = dnr;
    while (dnr_cur->next != NULL)
    {
      printf("%s\n", dnr_cur->path);
      temp = dnr_cur;
      dnr_cur = dnr_cur->next;
      free(temp);
    }
  }


  //print footer
  printf("\n\n");
  printf("Report generated by Adeos 1.0\n");
  printf("Mail bug reports or suggestions to <lamonml@linux.wku.edu>\n");
  printf("http://linux.wku.edu/~lamonml/software\n\n");
}


void writeHTMLHeaders()
{
  fp = fopen("results.html", "w");
  if (fp == NULL)
  {
    perror("results.html");
    exit(1);
  }

  fprintf(fp, "<HTML>\n");
  fprintf(fp, "<HEAD>\n");
  fprintf(fp, "<TITLE>Adeos Scan Results</TITLE>\n");
  fprintf(fp, "<style type=\"text/css\">\n");
  fprintf(fp, "<!--\n");
  fprintf(fp, ".BodyText {font-size:10pt; font-family:verdana,arial,helvetica;}\n");
  fprintf(fp, ".TitleText {font-size:24pt; font-family:verdana,arial,helvetica;}\n");
  fprintf(fp, "-->\n");
  fprintf(fp, "</style>\n");
  fprintf(fp, "</HEAD>\n");
  fprintf(fp, "<BODY>\n");
  fprintf(fp, "<B CLASS=TitleText>Adeos Scan Results</B>\n");
  fprintf(fp, "<HR ALIGN=LEFT SIZE=1 NOSHADE WIDTH=\"400\">\n");
  fprintf(fp, "<P>\n");
  
  if (!report)
  {
    fprintf(fp, "<TABLE BORDER=1 CELLPADDING=0 CELLSPACING=0 CLASS=BodyText>\n");
    fprintf(fp, "<TR>\n");
    fprintf(fp, "<TD WIDTH=200><CENTER><B>Status Message</B></CENTER></TD>\n");
    fprintf(fp, "<TD><CENTER><B>File Name</B></CENTER></TD>\n");
    fprintf(fp, "</TR>\n");
  }
}


void writeHTMLFooters()
{
  if (!report)
    fprintf(fp, "</TABLE>\n");
  fprintf(fp, "<P CLASS=BodyText>\n");
  fprintf(fp, "Report generated by Adeos 1.0<BR>\n");
  fprintf(fp, "Mail bug reports or suggestions to <A HREF=mailto:lamonml@linux.wku.edu>lamonml@linux.wku.edu</A><BR>\n");
  fprintf(fp, "<A HREF=\"http://linux.wku.edu/~lamonml/software\">http://linux.wku.edu/~lamonml/software</A><BR>\n");
  fprintf(fp, "</BODY>\n");
  fprintf(fp, "</HTML>\n");
  fclose(fp);
}


void outputHTMLReport()
{
  //output SUID file list
  fprintf(fp, "<P>\n");
  fprintf(fp, "<CENTER>\n");
  fprintf(fp, "<TABLE WIDTH=600 BORDER=1 CELLPADDING=0 CELLSPACING=0 CLASS=BodyText>\n");
  fprintf(fp, "<TR>\n");
  fprintf(fp, "<TD BGCOLOR=red><CENTER><B>SUID Files</B></CENTER></TD>\n");
  fprintf(fp, "</TR>\n");
  s_isuid_cur = s_isuid;
  while (s_isuid_cur->next != NULL)
  {
    fprintf(fp, "<TR>\n");
    fprintf(fp, "<TD>%s</TD>\n", s_isuid_cur->path);
    fprintf(fp, "</TR>\n");
    temp = s_isuid_cur;
    s_isuid_cur = s_isuid_cur->next;
    free(temp);
  }
  fprintf(fp, "</TABLE>\n");
  fprintf(fp, "</CENTER>\n");
  fprintf(fp, "<P>\n");

  //output SGID file list
  fprintf(fp, "<P>\n");
  fprintf(fp, "<CENTER>\n");
  fprintf(fp, "<TABLE WIDTH=600 BORDER=1 CELLPADDING=0 CELLSPACING=0 CLASS=BodyText>\n");
  fprintf(fp, "<TR>\n");
  fprintf(fp, "<TD BGCOLOR=red><CENTER><B>SGID Files</B></CENTER></TD>\n");
  fprintf(fp, "</TR>\n");
  s_isgid_cur = s_isgid;
  while (s_isgid_cur->next != NULL)
  {
    fprintf(fp, "<TR>\n");
    fprintf(fp, "<TD>%s</TD>\n", s_isgid_cur->path);
    fprintf(fp, "</TR>\n");
    temp = s_isgid_cur;
    s_isgid_cur = s_isgid_cur->next;
    free(temp);
  }
  fprintf(fp, "</TABLE>\n");
  fprintf(fp, "</CENTER>\n");
  fprintf(fp, "<P>\n");

  //output "sticky" files
  if (verbose || paranoid)
  {
    fprintf(fp, "<P>\n");
    fprintf(fp, "<CENTER>\n");
    fprintf(fp, "<TABLE WIDTH=600 BORDER=1 CELLPADDING=0 CELLSPACING=0 CLASS=BodyText>\n");
    fprintf(fp, "<TR>\n");
    fprintf(fp, "<TD BGCOLOR=yellow><CENTER><B>\"Sticky\" Files</B></CENTER></TD>\n");
    fprintf(fp, "</TR>\n");
    s_isvtx_cur = s_isvtx;
    while (s_isvtx_cur->next != NULL)
    {
      fprintf(fp, "<TR>\n");
      fprintf(fp, "<TD>%s</TD>\n", s_isvtx_cur->path);
      fprintf(fp, "</TR>\n");
      temp = s_isvtx_cur->next;
      s_isvtx_cur = s_isvtx_cur->next;
      free(temp);
    }
  }
  fprintf(fp, "</TABLE>\n");
  fprintf(fp, "</CENTER>\n");
  fprintf(fp, "<P>\n");

  //output inaccessible files
  if (verbose || paranoid)
  {
    fprintf(fp, "<P>\n");
    fprintf(fp, "<CENTER>\n");
    fprintf(fp, "<TABLE WIDTH=600 BORDER=1 CELLPADDING=0 CELLSPACING=0 CLASS=BodyText>\n");
    fprintf(fp, "<TR>\n");
    fprintf(fp, "<TD BGCOLOR=red><CENTER><B>Inaccessible Files</B></CENTER></TD>\n");
    fprintf(fp, "</TR>\n");
    ns_cur = ns;
    while (ns_cur->next != NULL)
    {
      fprintf(fp, "<TR>\n");
      fprintf(fp, "<TD>%s</TD>\n", ns_cur->path);
      fprintf(fp, "</TR>\n");
      temp = ns_cur->next;
      ns_cur = ns_cur->next;
      free(temp);
    }
  }
  fprintf(fp, "</TABLE>\n");
  fprintf(fp, "</CENTER>\n");
  fprintf(fp, "<P>\n");

  //output world-readable files
  if (paranoid)
  {
    fprintf(fp, "<P>\n");
    fprintf(fp, "<CENTER>\n");
    fprintf(fp, "<TABLE WIDTH=600 BORDER=1 CELLPADDING=0 CELLSPACING=0 CLASS=BodyText>\n");
    fprintf(fp, "<TR>\n");
    fprintf(fp, "<TD><CENTER><B>World-Readable Files</B></CENTER></TD>\n");
    fprintf(fp, "</TR>\n");
    s_iroth_cur = s_iroth;
    while (s_iroth_cur->next != NULL)
    {
      fprintf(fp, "<TR>\n");
      fprintf(fp, "<TD>%s</TD>\n", s_iroth_cur->path);
      fprintf(fp, "</TR>\n");
      temp = s_iroth_cur->next;
      s_iroth_cur = s_iroth_cur->next;
      free(temp);
    }
  }
  fprintf(fp, "</TABLE>\n");
  fprintf(fp, "</CENTER>\n");
  fprintf(fp, "<P>\n");

  //output world-writeable file list
  fprintf(fp, "<P>\n");
  fprintf(fp, "<CENTER>\n");
  fprintf(fp, "<TABLE WIDTH=600 BORDER=1 CELLPADDING=0 CELLSPACING=0 CLASS=BodyText>\n");
  fprintf(fp, "<TR>\n");
  fprintf(fp, "<TD BGCOLOR=red><CENTER><B>World-Writeable Files</B></CENTER></TD>\n");
  fprintf(fp, "</TR>\n");
  s_iwoth_cur = s_iwoth;
  while (s_iwoth_cur->next != NULL)
  {
    fprintf(fp, "<TR>\n");
    fprintf(fp, "<TD>%s</TD>\n", s_iwoth_cur->path);
    fprintf(fp, "</TR>\n");
    temp = s_iwoth_cur;
    s_iwoth_cur = s_iwoth_cur->next;
    free(temp);
  }
  fprintf(fp, "</TABLE>\n");
  fprintf(fp, "</CENTER>\n");
  fprintf(fp, "<P>\n");

  //output world-executable files
  if (paranoid)
  {
    fprintf(fp, "<P>\n");
    fprintf(fp, "<CENTER>\n");
    fprintf(fp, "<TABLE WIDTH=600 BORDER=1 CELLPADDING=0 CELLSPACING=0 CLASS=BodyText>\n");
    fprintf(fp, "<TR>\n");
    fprintf(fp, "<TD><CENTER><B>World-Executable Files</B></CENTER></TD>\n");
    fprintf(fp, "</TR>\n");
    s_ixoth_cur = s_ixoth;
    while (s_ixoth_cur->next != NULL)
    {
      fprintf(fp, "<TR>\n");
      fprintf(fp, "<TD>%s</TD>\n", s_ixoth_cur->path);
      fprintf(fp, "</TR>\n");
      temp = s_ixoth_cur->next;
      s_ixoth_cur = s_ixoth_cur->next;
      free(temp);
    }
  }
  fprintf(fp, "</TABLE>\n");
  fprintf(fp, "</CENTER>\n");
  fprintf(fp, "<P>\n");

  //output world-readable directory list
  if (paranoid)
  {
    fprintf(fp, "<P>\n");
    fprintf(fp, "<CENTER>\n");
    fprintf(fp, "<TABLE WIDTH=600 BORDER=1 CELLPADDING=0 CELLSPACING=0 CLASS=BodyText>\n");
    fprintf(fp, "<TR>\n");
    fprintf(fp, "<TD><CENTER><B>World-Readable Directories</B></CENTER></TD>\n");
    fprintf(fp, "</TR>\n");
    ds_iroth_cur = ds_iroth;
    while (ds_iroth_cur->next != NULL)
    {
      fprintf(fp, "<TR>\n");
      fprintf(fp, "<TD>%s</TD>\n",ds_iroth_cur->path);
      fprintf(fp, "</TR>\n");
      temp = ds_iroth_cur->next;
      ds_iroth_cur = ds_iroth_cur->next;
      free(temp);
    }
  }
  fprintf(fp, "</TABLE>\n");
  fprintf(fp, "</CENTER>\n");
  fprintf(fp, "<P>\n");

  //output world-writeable directory list
  fprintf(fp, "<P>\n");
  fprintf(fp, "<CENTER>\n");
  fprintf(fp, "<TABLE WIDTH=600 BORDER=1 CELLPADDING=0 CELLSPACING=0 CLASS=BodyText>\n");
  fprintf(fp, "<TR>\n");
  fprintf(fp, "<TD BGCOLOR=red><CENTER><B>World-Writeable Directories</B></CENTER></TD>\n");
  fprintf(fp, "</TR>\n");
  ds_iwoth_cur = ds_iwoth;
  while (ds_iwoth_cur->next != NULL)
  {
    fprintf(fp, "<TR>\n");
    fprintf(fp, "<TD>%s</TD>\n", ds_iwoth_cur->path);
    fprintf(fp, "</TR>\n");
    temp = ds_iwoth_cur;
    ds_iwoth_cur = ds_iwoth_cur->next;
    free(temp);
  }
  fprintf(fp, "</TABLE>\n");
  fprintf(fp, "</CENTER>\n");
  fprintf(fp, "<P>\n");

  //output world-executable directory list
  if (paranoid)
  {
    fprintf(fp, "<P>\n");
    fprintf(fp, "<CENTER>\n");
    fprintf(fp, "<TABLE WIDTH=600 BORDER=1 CELLPADDING=0 CELLSPACING=0 CLASS=BodyText>\n");
    fprintf(fp, "<TR>\n");
    fprintf(fp, "<TD><CENTER><B>World-Executable Directories</B></CENTER></TD>\n");
    fprintf(fp, "</TR>\n");
    ds_ixoth_cur = ds_ixoth;
    while (ds_ixoth_cur->next != NULL)
    {
      fprintf(fp, "<TR>\n");
      fprintf(fp, "<TD>%s</TD>\n", ds_ixoth_cur->path);
      fprintf(fp, "</TR>\n");
      temp = ds_ixoth_cur;
      ds_ixoth_cur = ds_ixoth_cur->next;
      free(temp);
    }
  }
  fprintf(fp, "</TABLE>\n");
  fprintf(fp, "</CENTER>\n");
  fprintf(fp, "<P>\n");

  //output inaccessible directories
  if (verbose || paranoid)
  {
    fprintf(fp, "<P>\n");
    fprintf(fp, "<CENTER>\n");
    fprintf(fp, "<TABLE WIDTH=600 BORDER=1 CELLPADDING=0 CELLSPACING=0 CLASS=BodyText>\n");
    fprintf(fp, "<TR>\n");
    fprintf(fp, "<TD><CENTER><B>Inaccessible Directories</B></CENTER></TD>\n");
    fprintf(fp, "</TR>\n");
    dnr_cur = dnr;
    while (dnr_cur->next != NULL)
    {
      fprintf(fp, "<TR>\n");
      fprintf(fp, "<TD>%s</TD>\n", dnr_cur->path);
      fprintf(fp, "</TR>\n");
      temp = dnr_cur;
      dnr_cur = dnr_cur->next;
      free(temp);
    }
  }
  fprintf(fp, "</TABLE>\n");
  fprintf(fp, "</CENTER>\n");
  fprintf(fp, "<P>\n");
}

